#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
from scipy import special
from copy import deepcopy
import lmfit
import bindensity as bind
from pathos.multiprocessing import Pool
from ..ANTARESS_analysis.ANTARESS_inst_resp import convol_prof,conv_st_prof_tab
from ..ANTARESS_general.utils import closest,init_parallel_func,np_where1D,stop

##################################################################################################    
#%%% Model line profiles
##################################################################################################

def gen_fit_prof(param_in,x_tab,args=None):
    r"""**Generic profile fit function**
    
    Function called by minimization algorithms, returning the chosen model profile after applyying (if requested) convolution, spectral conversion and resampling.
    
    The profile model function must return several arguments (or a list of one) with the model as first argument.
    The profile is calculated over a continuous table to allow for convolution and use of covariance matrix (fitted pixels are accounted for directly in the minimization routine).

    Args:
        TBD
    
    Returns:
        TBD
    
    """      
    args_exp = deepcopy(args['args_exp'])

    #In case param_in is defined as a Parameters structure, retrieve values and define dictionary
    if isinstance(param_in,lmfit.parameter.Parameters):
        param={}
        for par in param_in:param[par]=param_in[par].value
    else:param=param_in   
   
   	#Profile model
    sp_line_model = args['func_nam'](param,args_exp['cen_bins'],args)[0]
   
   	#Convolution and resampling 
    mod_prof = conv_st_prof_tab(None,None,None,args,args_exp,sp_line_model,args['FWHM_inst'])  

    return mod_prof




def dispatch_func_prof(func_name):
    r"""**Line profile model dispatching**
    
    Returns the chosen line profile function.

    Args:
        func_name (str): name of the chosen function
    
    Returns:
        func (function): chosen function
    
    """  
    func = {'voigt':voigt,
            'gauss':gauss_intr_prof,
            'cgauss':gauss_herm_lin,
            'pgauss':gauss_poly,
            'dgauss':dgauss,
        }[func_name]
    return func





def voigt(param,x,args=None):
    r"""**Line model: Voigt**
    
    Calculates Voigt profile, expressed in terms of the Faddeeva function (factorless)

    .. math::     
       V(rv) = Real(w(x))     
           
    As a function of 
    
    .. math:: 
       x &= ((rv-rv_0) + i \gamma)/(\sqrt{2} \sigma)   \\
         &= (rv-rv_0)/(\sqrt{2} \sigma) + i a       \\
         &= 2 \sqrt{\ln{2}} (rv-rv_0)/\mathrm{FWHM}_\mathrm{gauss} + i a
       
    Where `a` is the damping coefficient 
    
    .. math:: 
       a &= \gamma/(\sqrt{2} \sigma)            \\
         &= 2 \gamma \sqrt{\ln{2}}/\mathrm{FWHM}_\mathrm{gauss}       \\
         &= \sqrt{\ln{2}} \mathrm{FWHM}_\mathrm{lor}/\mathrm{FWHM}_\mathrm{gauss}       
       
    The :math:`\mathrm{FWHM}_\mathrm{gauss}` parameter is that of the Gaussian component (factorless)
    
    .. math::
       G(rv) &= \exp(- (rv-rv_0)^2/ ( \mathrm{FWHM}_\mathrm{gauss}/(2 \sqrt{\ln{2}}) )^2  )            \\     
             &= \exp(- (rv-rv_0)^2/ 2 ( \mathrm{FWHM}_\mathrm{gauss}/(2 \sqrt{2 \ln{2}}) )^2  )         \\
             &= \exp(- (rv-rv_0)^2/ 2 \sigma^2  )

    With :math:`\sigma = \mathrm{FWHM}_\mathrm{gauss}/(2 \sqrt{\ln{2}})` and :math:`\mathrm{FWHM}_\mathrm{lor} = 2 \gamma`
   
    The full-width at half maximum of the Voigt profile approximates as

    .. math::    
       \mathrm{FWHM}_\mathrm{voigt} = 0.5436 \mathrm{FWHM}_\mathrm{lor}+ \sqrt{0.2166 \mathrm{FWHM}_\mathrm{lor}^2 + \mathrm{FWHM}_\mathrm{gauss}^2} 
    
    From J.J.Olivero and R.L. Longbothum in Empirical fits to the Voigt line width: A brief review, JQSRT 17, P233    

    Args:
        TBD
    
    Returns:
        TBD
    
    """  
    z_tab = 2.*np.sqrt(np.log(2.))*(x - param['rv' ])/param['FWHM'] +  1j*param['a_damp']
    
    #Voigt profile
    voigt_peak = special.wofz(1j*param['a_damp']).real
    voigt_mod = 1. - (param['ctrst']/voigt_peak)*special.wofz(z_tab).real
  
    #Continuum        
    cont_pol = param['cont']*pol_cont(x,args,param)   

    return voigt_mod*cont_pol , cont_pol



def gauss_intr_prof(param,rv_tab,args=None):
    r"""**Line model: Gaussian**
    
    Calculates Gaussian profile.
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """ 
    ymodel = [param['cont']*(1.-param['ctrst']*np.exp(-np.power( 2.*np.sqrt(np.log(2.))*(rv_tab-param['rv'])/param['FWHM']  ,2.  )))] 
    return ymodel





def gauss_herm_lin(param,x,args=None):
    r"""**Line model: Skewed Gaussian**
    
    Calculates Gaussian profile with hermitian polynom term for skewness/kurtosis.
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """    
    x_tab = (2.*np.sqrt(2.*np.log(2.)))*(x-param['rv' ])/param['FWHM'] 
    
    #Skewness and kurtosis
    skew1 = param['skewA']
    kurt1 = param['kurtA']
    factor=1.
    if skew1!=0. or kurt1!=0.:  
        c0 = np.sqrt(6.)/4.
        c1 = -np.sqrt(3.)
        c2 = -np.sqrt(6.)
        c3 = 2./np.sqrt(3.)
        c4 = np.sqrt(6.)/3.        
    if skew1!=0.:factor+=skew1*(c1*x_tab+c3*x_tab**3.)
    if kurt1!=0.:factor+=kurt1*(c0+c2*x_tab**2.+c4*x_tab**4.)   
 
    #Skewd gaussian profile
    sk_gauss = 1.- factor*param['ctrst']*np.exp(-0.5*x_tab**2.)
     
    #Continuum        
    cont_pol = param['cont']*pol_cont(x,args,param)    

    return sk_gauss*cont_pol , cont_pol



def gauss_poly(param,RV,args=None):
    r"""**Line model: Sidelobed Gaussian**
    
    Calculates Gaussian profile with flat continuum and polynomial for sidelobes.
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """ 
 
    #Gaussian with baseline set to continuum value
    cen_RV = param['rv']    #center
    y_gausscore=1.-param['ctrst']*np.exp(-np.power( 2.*np.sqrt(np.log(2.))*(RV-cen_RV)/param['FWHM']  ,2.  )) 
    ymodel = y_gausscore*param['cont']
    
    #Polynomial
    #    - P(x) = a0 + a1*(x-x0) + a2*(x-x0)^2 + a3*(x-x0)^3 + a4*(x-x0)^4 + a6*(x-x0)^6
    #      P must by symmetric, and have continuity of value and derivative in x=x0+-dx
    #      this must be true in particular if x0 = 0, and we thus solve:
    # P(x) = P(-x)
    # P(dx) = cont
    # P'(dx) = 0 
    #    - the continuum is set to its constant value beyond the continuity points
    c4_pol = param['c4_pol']                          #4th order coefficient
    c6_pol = param['c6_pol']                          #6th order coefficient
    dRV_joint=param['dRV_joint']                      #continuity points
    RV_joint_high = cen_RV + dRV_joint       
    RV_joint_low  = cen_RV - dRV_joint
    cond_lobes = (RV >= RV_joint_low) & (RV <= RV_joint_high) 
    y_polylobe = np.repeat(1.,len(RV))
    y_polylobe[cond_lobes] *= c4_pol*dRV_joint**4. + 2.*c6_pol*dRV_joint**6. - dRV_joint**2.*np.power(RV[cond_lobes]-cen_RV,2.)*(2.*c4_pol + 3.*c6_pol*dRV_joint**2.) + c4_pol*np.power(RV[cond_lobes]-cen_RV,4.) + c6_pol*np.power(RV[cond_lobes]-cen_RV,6.)
    ymodel[cond_lobes]*=y_polylobe[cond_lobes]    
    
    return ymodel, param['cont']*y_gausscore, param['cont']*y_polylobe
                 


def dgauss(param,rv_tab,args=None):
    r"""**Line model: Double Gaussian**
    
    Calculates Double-Gaussian profile, with Gaussian in absorption co-added with a Gaussian in emission at the core.
    
    The Gaussian profiles are centered but their amplitude and width can be fixed or independant. The model is defined as

    .. math::    
       F &= F_\mathrm{cont} + A_\mathrm{core} \exp(f_1) + A_\mathrm{lobe} \exp(f_2)       \\
       F &= F_\mathrm{cont} + A_\mathrm{core} ( \exp(f_1) - A_\mathrm{l2c} \exp(f_2) ) 
       
    We define a contrast parameter as the relative flux between the continuum and the CCF minimum 

    .. math::    
       C &= (F_\mathrm{cont} - (F_\mathrm{cont} + A_\mathrm{core} - A_\mathrm{core} A_\mathrm{l2c}) ) / F_\mathrm{cont}          \\
       C &= -A_\mathrm{core} ( 1 - A_\mathrm{l2c}) / F_\mathrm{cont}              \\
       A_\mathrm{core} &= -C F_\mathrm{cont}/( 1 - A_\mathrm{l2c})             \\
       A_\mathrm{lobe} &= -A_\mathrm{l2c} A_\mathrm{core} = A_\mathrm{l2c} C F_\mathrm{cont}/( 1 - A_\mathrm{l2c})
      
    Thus the model can be expressed as   

    .. math::     
       F &= F_\mathrm{cont} - C F_\mathrm{cont} ( \exp(f_1) - A_\mathrm{l2c} \exp(f_2) )/( 1 - A_\mathrm{l2c})             \\
       F &= F_\mathrm{cont} ( 1 - C ( \exp(f_1) - A_\mathrm{l2c} \exp(f_2) )/( 1 - A_\mathrm{l2c}) )     
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """ 
 
    #Reduced contrast
    ctrst_red = param['ctrst']/( 1. - param['amp_l2c'])
     
    #Inverted gaussian core
    y_gausscore=param['cont']*(1. - ctrst_red*np.exp(-np.power(2.*np.sqrt(np.log(2.))*(rv_tab-param['rv'])/param['FWHM'],2))) 
    
    #Gaussian lobes
    cen_RV_lobes=param['rv']+param['rv_l2c']   
    FWHM_lobes = (param['FWHM'])*(param['FWHM_l2c'])   
    y_gausslobes=param['cont']*(1. + ctrst_red*param['amp_l2c']*np.exp(-np.power(2.*np.sqrt(np.log(2.))*(rv_tab-cen_RV_lobes)/FWHM_lobes,2)))     
    
    #Complete model
    ymodel = y_gausscore + y_gausslobes - param['cont']

    #Output models of each gaussian component superimposed to the continuum
    return ymodel , y_gausscore, y_gausslobes






def pol_cont(cen_bins_ref,args,param):
    r"""**Line polynomial continuum**
    
    Calculates polynomial modulated around 1.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    return (1. + param['c1_pol']*cen_bins_ref + param['c2_pol']*cen_bins_ref**2.  + param['c3_pol']*cen_bins_ref**3. + param['c4_pol']*cen_bins_ref**4. ) 



def calc_macro_ker_rt(rv_mac_kernel,param,cos_th,sin_th):
    r"""**Macroturbulence broadening: anisotropic Gaussian**
    
    Calculates broadening kernel for local macroturbulence, based on formulation with anisotropic gaussian (Takeda \& UeNo 2017)

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    
    if cos_th>0:term_R=(param['A_R']/(np.sqrt(np.pi)*param['ksi_R']*cos_th))*np.exp(-rv_mac_kernel**2./  ((param['ksi_R']*cos_th)**2.)  )   #Otherwise exp yield 0
    else:term_R=0.
    if sin_th>0:term_T=(param['A_T']/(np.sqrt(np.pi)*param['ksi_T']*sin_th))*np.exp(-rv_mac_kernel**2./  ((param['ksi_T']*sin_th)**2.)  )   #Otherwise exp yield 0 
    else:term_T=0.
    macro_ker_sub=term_R+term_T    
    return macro_ker_sub

def calc_macro_ker_anigauss(rv_mac_kernel,param,cos_th,sin_th):
    r"""**Macroturbulence broadening: radial-tangential**
    
    Calculates broadening kernel for local macroturbulence, based on radial-tangential model (Gray 1975, 2005)

    Args:
        TBD
    
    Returns:
        TBD
    
    """     
    macro_ker_sub=np.exp(  -rv_mac_kernel**2./  ((param['eta_R']*cos_th)**2.  +  (param['eta_T']*sin_th)**2.)  )
    return macro_ker_sub





##################################################################################################    
#%%% Model line properties
##################################################################################################

def calc_linevar_coord_grid(dim,grid):
    r"""**Line profile coordinate dispatching**
    
    Returns the relevant coordinate for calculation of line properties variations.

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 
    if (dim in ['mu','xp_abs','r_proj','y_st']):linevar_coord_grid = grid[dim]
    elif (dim=='y_st2'):linevar_coord_grid = grid['y_st']**2.
    elif (dim=='abs_y_st'):linevar_coord_grid = np.abs(grid['y_st'])
    else:stop('Undefined line coordinate')
    return linevar_coord_grid


def calc_polymodu(pol_mode,coeff_pol,x_val):
    r"""**Line profile coordinate models**
    
    Calculates absolute or modulated polynomial
    
    The list of polynomial coefficient 'coeff_pol' has been defined in decreasing powers, as expected by `poly1d`, using input coefficient defined through their power value (see `polycoeff_def()`)

    Args:
        TBD
    
    Returns:
        TBD
    
    """     
    
    #Absolute polynomial
    #    - coeff_pol[0]*x^n + coeff_pol[1]*x^(n-1) .. + coeff_pol[n]*x^0
    if pol_mode=='abs':
        mod= np.poly1d(coeff_pol)(x_val)         

    #Modulated polynomial
    #    - (coeff_pol[0]*x^n + coeff_pol[1]*x^(n-1) .. + 1)*coeff_pol[n] 
    elif pol_mode=='modul':
        coeff_pol_modu = coeff_pol[:-1] + [1]
        mod= coeff_pol[-1]*np.poly1d(coeff_pol_modu)(x_val)  
    else:stop('Undefined polynomial mode')
    return mod





def polycoeff_def(param,coeff_ord2name_polpar):
    r"""**Line profile coordinate coefficients**
    
    Defines polynomial coefficients from the `Parameter()` format through their power value

    Args:
        TBD
    
    Returns:
        TBD
    
    """ 

    #Polynomial coefficients 
    #    - keys in 'coeff_ord2name_polpar' are the coefficients degrees, values are their names, as defined in 'param' 
    #      they can be defined in disorder (in terms of degrees), as coefficients are forced to order from deg_max to 0 in 'coeff_grid_polpar' 
    #    - degrees can be missing
    #    - input coefficients must be given in decreasing order of degree to poly1d
    deg_max=max(coeff_ord2name_polpar.keys())
    coeff_grid_polpar=[param[coeff_ord2name_polpar[ideg]] if ideg in coeff_ord2name_polpar else 0. for ideg in range(deg_max,-1,-1)]

    return coeff_grid_polpar





# Stage Théo : fusion des 2 dernières fonctions. 
def poly_prop_calc(param,fit_coord_grid,coeff_ord2name_polpar, pol_mode):

    #Polynomial coefficients 
    #    - keys in 'coeff_ord2name_polpar' are the coefficients degrees, values are their names, as defined in 'param' 
    #      they can be defined in disorder (in terms of degrees), as coefficients are forced to order from 0 to deg_max in 'coeff_pol_ctrst' 
    #    - degrees can be missing
    #     - input coefficients must be given in decreasing order of degree to poly1d
    deg_max=max(coeff_ord2name_polpar.keys())
    
    coeff_grid_polpar=[param[coeff_ord2name_polpar[ideg]] if ideg in coeff_ord2name_polpar else 0. for ideg in range(deg_max,-1,-1)]
    
    # Absolute polynomial 
    if pol_mode == 'abs' : 
        polpar_grid = np.poly1d(coeff_grid_polpar)(fit_coord_grid)
        
    # Modulated polynomial
    if pol_mode == 'modul' :
        coeff_pol_modu = coeff_grid_polpar[:-1] + [1]
        polpar_grid = np.poly1d(coeff_pol_modu)(fit_coord_grid)*coeff_grid_polpar[-1]
        
    return polpar_grid









##################################################################################################    
#%%% Model line analysis
##################################################################################################

def calc_biss(Fnorm_in,RV_tab_in,RV_min,max_rv_range,dF_grid,resamp_mode,Cspan):
    r"""**Bissector.**
    
    Calculates bissector for a line profile provided in RV space as input.
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """ 
    #Reduction to maximum range    
    cond_range = (RV_tab_in>max_rv_range[0]) & (RV_tab_in<max_rv_range[-1])
    Fnorm = Fnorm_in[cond_range]
    RV_tab = RV_tab_in[cond_range]

    #Blue and red wings of input line
    idx_blue=np_where1D(RV_tab<RV_min)
    idx_red=np_where1D(RV_tab>=RV_min)
  
    #Calculation range
    #    - between the maxima on each side of the line
    max_blue=max(Fnorm[idx_blue])
    max_red=max(Fnorm[idx_red])
    idx_max_blue=np_where1D(Fnorm==max_blue)[0]
    idx_max_red=np_where1D(Fnorm==max_red)[0]
    idx_min = closest(RV_tab,RV_min)
    idx_biss_blue=range(idx_max_blue,idx_min+1)
    idx_biss_red=range(idx_min,idx_max_red+1)
    Fnorm_biss = Fnorm[idx_max_blue:idx_max_red+1]
     
    #RV = f(flux) in blue and red wings
    RV_biss_blue=RV_tab[idx_biss_blue]
    F_biss_blue=Fnorm[idx_biss_blue]
    idx_sort = np.argsort(F_biss_blue)
    RV_biss_blue = RV_biss_blue[idx_sort]
    F_biss_blue = F_biss_blue[idx_sort]
    Fedge_biss_blue = 0.5*(F_biss_blue[0:-1]+F_biss_blue[1:])
    Fedge_biss_blue = np.concatenate(( [F_biss_blue[0] - 0.5*(F_biss_blue[1]-F_biss_blue[0])] , Fedge_biss_blue , [F_biss_blue[-1]+ 0.5*(F_biss_blue[-1]-F_biss_blue[-2])] ))
    RV_biss_red=RV_tab[idx_biss_red]
    F_biss_red=Fnorm[idx_biss_red]
    idx_sort = np.argsort(F_biss_red)
    RV_biss_red = RV_biss_red[idx_sort]
    F_biss_red = F_biss_red[idx_sort]
    Fedge_biss_red = 0.5*(F_biss_red[0:-1]+F_biss_red[1:])
    Fedge_biss_red = np.concatenate(([ F_biss_red[0] - 0.5*(F_biss_red[1]-F_biss_red[0])] , Fedge_biss_red , [F_biss_red[-1]+ 0.5*(F_biss_red[-1]-F_biss_red[-2])] ))
    
    #Resampling RV versus flux
    minF_biss=np.min(Fnorm_biss)
    maxF_biss=np.max(Fnorm_biss)
    nF_biss=int((maxF_biss-minF_biss)/dF_grid)
    F_biss_edgeHR=minF_biss+dF_grid*np.arange(nF_biss+1)
    RV_biss_blue_HR = bind.resampling(F_biss_edgeHR,Fedge_biss_blue,RV_biss_blue,kind = resamp_mode)            
    RV_biss_red_HR = bind.resampling(F_biss_edgeHR,Fedge_biss_red,RV_biss_red,kind = resamp_mode)

    #Bissector
    RV_biss_HR=0.5*(RV_biss_blue_HR+RV_biss_red_HR)
    F_biss_HR = 0.5*(F_biss_edgeHR[0:-1]+F_biss_edgeHR[1:])
    
    #Bissector span
    #    - difference in velocity between minimum and region of larger variation in the series
    if Cspan is not None:idx_span = closest(F_biss_HR,1. - Cspan*(1.-minF_biss) )
    else:idx_span = np.argmax(np.abs(RV_biss_HR-RV_min))
    RV_biss_span = RV_biss_HR[idx_span]
    F_biss_span = F_biss_HR[idx_span]
    bis_span = RV_biss_span - RV_min
                                  
    return RV_biss_HR,F_biss_HR,RV_biss_span,F_biss_span,bis_span



def gauss_intr_prop(ctrst_intr,FWHM_intr,FWHM_inst):
    r"""**Intrinsic > Measured Gaussian properties.**
    
    Estimates the FWHM and contrast of a measured-like Gaussian profile, defined as the convolution of a Gaussian intrinsic model profile and a Gaussian instrumental response.

    The FWHM can be expressed as
    
    .. math:: 
        \mathrm{FWHM} &= \sqrt{\mathrm{FWHM}_\mathrm{intr}^2 + \mathrm{FWHM}_\mathrm{inst}^2 } \\
        \mathrm{with}\,\sigma &= \frac{\mathrm{FWHM}}{2 \sqrt{\ln{2}}} = \sqrt{\sigma_\mathrm{intr}^2 + \sigma_\mathrm{inst}^2 }
        
    The area of the measured convolution profile is equal to the product of the area of the intrinsic and instrumental convolved profiles, providing an analytical expression for the contrast.  
    The instrumental kernel, with area = 1, is expressed as

    .. math::     
        f_\mathrm{kern}(rv) &= \frac{1}{\sqrt{2 \pi} \sigma_\mathrm{inst} } \exp^{ -\frac{rv^2}{2 \sigma_\mathrm{inst}^2} } \\
        \mathrm{where}\,\sigma_\mathrm{inst} &= \frac{\mathrm{FWHM}_\mathrm{inst}}{ 2 \sqrt{2 \log{2}} }
            
    The intrinsic profile, with area = :math:`\mathrm{A}_\mathrm{intr}`, is expressed as
    
    .. math::     
        f_\mathrm{intr}(rv) &= \frac{\mathrm{A}_\mathrm{intr}}{\sqrt{2 \pi} \sigma_\mathrm{intr} } \exp^{ -\frac{rv^2}{2 \sigma_\mathrm{intr}^2} } \\
        \mathrm{where}\,\sigma_\mathrm{intr} &= \frac{\mathrm{FWHM}_\mathrm{intr}}{ 2 \sqrt{2 \ln{2}} }    
    
    The contrast `C` of the measured profile relates to its area `A` as
    
    .. math::     
        C &= \frac{A}{\sqrt{2 \pi} \sigma }  \\
          &= \frac{1 \times \mathrm{A}_\mathrm{intr} }{ \sqrt{2 \pi (\sigma_\mathrm{intr}^2 + \sigma_\mathrm{inst}^2 ) }}  \\
          &= \frac{\mathrm{C}_\mathrm{intr} \sqrt{2 \pi} \sigma_\mathrm{intr} }{\sqrt{2 \pi (\sigma_\mathrm{intr}^2 + \sigma_\mathrm{inst}^2 ) }}   \\        
          &= \frac{\mathrm{C}_\mathrm{intr} \sigma_\mathrm{intr} }{\sqrt{\sigma_\mathrm{intr}^2 + \sigma_\mathrm{inst}^2}}   \\          
          &= \frac{\mathrm{C}_\mathrm{intr} \mathrm{FWHM}_\mathrm{intr} }{\sqrt{\mathrm{FWHM}_\mathrm{intr}^2 + \mathrm{FWHM}_\mathrm{inst}^2}}    
                                     
    This is only valid for analytical profiles. For numerical profiles, the model is the sum of intrinsic profiles over the planet-occulted surface that may have non-Gaussian shape, and which
    are further broadened by the local surface `rv` field. In that case, use `cust_mod_true_prop()`. 
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """ 
    FWHM_meas = np.sqrt(FWHM_intr**2. + FWHM_inst**2.) 
    ctrst_meas = ctrst_intr*FWHM_intr/FWHM_meas
    return ctrst_meas,FWHM_meas







def cust_mod_true_prop(param,velccf_HR,args):
    r"""**Intrinsic > Measured line properties.**
    
    Estimates the FWHM and contrast of a measured-like line profile with custom shape.
    This is done numerically, calculating the model profile at high resolution.
    
    Contrast is counted from the estimated continuum.
    For double-gaussian profiles, the peaks of the lobes are considered as the closest estimate to the true stellar continuum.
    
    FWHM is calculated by finding the points on the blue side at half-maximum, and on the red side.
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """ 
    #HR model
    CCF_HR_intr = args['fit_func_gen'](param,velccf_HR,args=args)[0]
    
    #Instrumental convolution 
    if args['FWHM_inst'] is not None:CCF_HR = convol_prof(CCF_HR_intr,velccf_HR,args['FWHM_inst'])  
    else:CCF_HR=CCF_HR_intr
    
    #Approximation of continuum
    max_CCF = np.max(CCF_HR)
    
    #CCF inverted peak
    min_CCF = np.min(CCF_HR)
    
    #Estimates of amplitude and contrast
    true_amp=max_CCF - min_CCF 
    true_ctrst=1. - (min_CCF/max_CCF) 
    
    #Blue side and red side of model CCF
    blueRV_HR=np_where1D(velccf_HR<=param['rv'])
    if len(blueRV_HR)==0:blueRV_HR=[0]
    redRV_HR=np_where1D(velccf_HR>=param['rv'])
    if len(redRV_HR)==0:redRV_HR=[-1]
    
    #Model pixels closest to half-maximum (mid-point between maximum and minimum of model) on each wing
    idx_blue_WHM=closest(CCF_HR[blueRV_HR],0.5*(np.max(CCF_HR[blueRV_HR])+min_CCF))  
    idx_red_WHM=closest(CCF_HR[redRV_HR],0.5*(np.max(CCF_HR[redRV_HR])+min_CCF))   
    
    #Estimate of FWHM defined as sum of HWHM in each wing
    true_FWHM = (param['rv'] - velccf_HR[blueRV_HR][idx_blue_WHM]) + (velccf_HR[redRV_HR][idx_red_WHM] - param['rv'])       
           
    return true_ctrst,true_FWHM,true_amp    

def proc_cust_mod_true_prop(merged_chain_proc,args,param_loc):
    r"""**Wrap-up of cust_mod_true_prop().**
    
    Estimates the FWHM and contrast of a measured-like line profile with custom shape, for a sample of properties.
    
    Args:
        TBD
    
    Returns:
        TBD
        
    """     
    nsamp=len(merged_chain_proc[:,0])    
    param_proc = deepcopy(param_loc)
    chain_loc_proc=np.empty([3,0],dtype=float) 
    for istep in range(nsamp): 
        for ipar,par in enumerate(args['var_par_list']):param_proc[par]=merged_chain_proc[istep,ipar]     
        true_ctrst_step,true_FWHM_step,true_amp_step=cust_mod_true_prop(param_proc,args['cen_bins_HR'],args)  
        chain_loc_proc=np.append(chain_loc_proc,[[true_ctrst_step],[true_FWHM_step],[true_amp_step]],axis=1) 
    return chain_loc_proc
    
def para_cust_mod_true_prop(func_input,nthreads,n_elem,y_inputs,common_args):  
    r"""**Multithreading routine for proc_cust_mod_true_prop().**

    Args:
        func_input (function): multi-threaded function
        nthreads (int): number of threads
        n_elem (int): number of elements to thread
        y_inputs (list): threadable function inputs 
        common_args (tuple): common function inputs
    
    Returns:
        y_output (None or specific to func_input): function outputs 
    
    """        
    pool_proc = Pool(processes=nthreads)   #cannot be passed through lmfit
    ind_chunk_list=init_parallel_func(nthreads,n_elem)
    chunked_args=[(y_inputs[0][ind_chunk[0]:ind_chunk[1],:],)+common_args for ind_chunk in ind_chunk_list]	
    all_results=tuple(tab for tab in pool_proc.starmap(func_input,chunked_args))				
    y_output=np.concatenate(tuple(all_results[i] for i in range(nthreads)),axis=1)
    pool_proc.close()
    pool_proc.join() 
    return y_output



































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from math import pi
import numpy as np

#######################################################################
#A routine to define planet parameters and physical constants
#    - check regularly for papers with more accurate data
#######################################################################

#################################################################
#Astrophysical
#    - online catalog use the equatorial radii of Solar system planets as reference, as do probably most papers
#################################################################

AU_1=149597870.7    #(km) astronomical unit [IAU resolution 2012]
AU_1_m=AU_1*1e3     #(m)
AU_1_cm=AU_1_m*1e2  #(cm)
pc_1=648000./pi     #(au) parsec  [IAU resolution 2015]
c_light=299792.458  #(km/s) speed of light 
c_light_m=299792458.  #(m/s) speed of light
c_light_A=2.99792458e+18  #(A/s) speed of light
AUyr_kms=AU_1/(365.25*24.*3600.)
     
#Sun 
Msun=1.988415829e30 #(kg) solar mass [IAU resolution 2015, with G from CODATA 2006]  
Rsun=695700.        #(km) solar radius [IAU resolution 2015]
Lsun=382.8*1e24     #(kg⋅m2⋅s-3) solar luminosity [IAU resolution 2015]
    
#Jupiter : 
Mjup=1.898130285e27 #(kg) Jupiter mass [IAU resolution 2015, with G from CODATA 2006]  
Rjup=71492.         #(km) equatorial Jupiter radius [IAU resolution 2015] 
    
#Neptune : http://nssdc.gsfc.nasa.gov/planetary/factsheet/neptunefact.html
Mnept=102.42e24      #(kg)
Rnept=24764.         #(km) equatorial radius
    
#Earth : 
Mearth=5.972186e24 #(kg) Earth mass [IAU resolution 2015, with G from CODATA 2006]  
Rearth=6378.1      #(km) equatorial Earth radius [IAU resolution 2015] 

#################################################################
#Physical/chemical
#################################################################

G_usi=6.67428e-11        #(m3 kg-1 s-2) gravitational constant [CODATA 2006 value, see IAU resolution 2015]
sig_sb=5.670373*1e-8     #(kg s-3 K-4) Stefan-Boltzmann constant 
amu=1.660531e-27         #(kg) Atomic mass unit 
k_boltz=1.3806488*1e-23  #(m2 kg s-2 K-1) Boltzmann constant 
h_planck=6.62606957e-34  #(kg m2 s-1) Planck constant   
alpha=2.654008849418e-06 #(m2 s-1) = (pi e^2)/ (4 pi Epsilon_0 m_e c)  with e, epsilon_0, m_e, c in USI
                         #           e = 1.60217662e-19 A s; Eps_0 = 8.854187817620e−12 A^2 s^4 kg−1 m−3; m_e = 9.10938356e-31 kg 
N_avo=6.02214129e23      #(mol-1) = Avogadro constant
elec_ch=1.6021766208e-19 #(C = A s) = elementary charge (proton charge)

#Loschmidt's number (in cm-3) 
#    - number of molecules per unit of volume
#    - N_Loschmidt = N_Avo / V_mol
#           = N_Avo / (R * T / p) = p / (k_B * T) = density for a perfect gas  
#    - value at 0°C and 1 atm is (https://en.wikipedia.org/wiki/Loschmidt_constant):
# N_Loschmidt = 2.6867805e19 /cm3
#      thus value at T = 15°C and p = 1013 hPa (1 atm) is:
# N_Loschmidt[15°C] = N_Loschmidt[0] * T[0°C]/T[15°C]      
#                   = 2.6867805e19 * 273.15 / (273.15 + 15)
#                   = 2.5469168612701716e+19 /cm3 
N_Loschmidt = 2.5469168612701716e+19

#Pressure normalization coefficient
#    - P = n*k*T
# with P in Pa (kg m-1 s-2)  
#      n the number density in atoms/m3
#      k in m2 kg s-2 K-1
#      T in K
#    - conversion
# 1 Pa = 1e-5 bar -> P[bar] = P[Pa]*1e-5 
# 1 atm = 1.01325 bar
# 1 bar = 0.9869232667160128 atm -> P[atm] = P[bar]*0.9869232667160128 
#      thus:
# P[Pa]  = 1e6*n[atoms/cm3]*k[m2 kg s-2 K-1]*T[K]
# P[bar] = 1e6*n[atoms/cm3]*k[m2 kg s-2 K-1]*T[K]*1e-5
# P[atm] = 1e6*n[atoms/cm3]*k[m2 kg s-2 K-1]*T[K]*1e-5*0.9869232667160128 
# P[atm] = Patm_Ncm3_Tk*n[atoms/cm3]*T[K]
#      with Patm_Ncm3_Tk = 1e6*k[m2 kg s-2 K-1]*1e-5*0.9869232667160128
#                      = 9.869232667160128*k[m2 kg s-2 K-1]
#                      = 1.3625944238835428e-22
Patm_Ncm3_Tk=k_boltz*9.869232667160128



#################################################################
#Atomic & molecular data
#    - name is 'X_Y_Z', where
# + X is the element name
# + Y is its ionization level
# + Z is its excited level
#    - properties:
# + m_at : mass of the atom/ion/molecule (in atomic mass unit)
# masses can be retrieved from http://webbook.nist.gov/chemistry/name-ser.html
# or https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl
# + sig_coll : collision cross section between two of the species particles (in m^2) 
#                 if unknown, set to pi*(2*r)^2
# + ion_thresh: ionization threshold for the species (A)
# + q: electric charge (in elementary charges)
#################################################################

dico_all_gases={
    
    #-------------------------
    #Atoms & ion
    #-------------------------
    
    #Hydrogen
    'H_0+_g':{'nature':'gas','m_at':1.007975,'sig_coll':1e-21,'ion_thresh':911.8,'q':0.}, 
    'H_1+_g':{'nature':'gas','m_at':1.00739,'q':elec_ch}, 

    #Deuterium
    'D_0+_g':{'nature':'gas','m_at':2.0141017778,'q':0.},

    #Helium
    #    - radius = 31 pm 
    #    - collisionnal cross-section set to pi*(2.*31.*1e-12)**2.
    'He_0+_g':{'nature':'gas','m_at':4.002602,'ion_thresh':504.2055460784134,'sig_coll':1.208e-20,'q':0.}, 
    'He_0+_23S1':{'nature':'gas','m_at':4.002602,'ion_thresh':2593.01,'sig_coll':1.208e-20}, 
    
    #Carbon
    'C_0+_g':{'nature':'gas','m_at':12.0107,'ion_thresh':1101.1025202547235,'q':0.},
    'C_1+_g':{'nature':'gas','m_at':12.0107,'ion_thresh':508.5485799043555,'q':elec_ch},
    'C_2+_g':{'nature':'gas','m_at':12.0107,'q':2.*elec_ch},

    #Nitrogen
    'N_0+_g':{'nature':'gas','m_at':14.007,'ion_thresh':853.297617210474,'q':0.},
    'N_1+_g':{'nature':'gas','m_at':14.007,'ion_thresh':418.86535061041167,'q':elec_ch},
    'N_2+_g':{'nature':'gas','m_at':14.007,'ion_thresh':261.2942966926909,'q':2.*elec_ch}, 
    'N_3+_g':{'nature':'gas','m_at':14.007,'ion_thresh':160.04149190742464,'q':3.*elec_ch}, 
    'N_4+_g':{'nature':'gas','m_at':14.007,'ion_thresh':126.65659799844914,'q':4.*elec_ch}, 
    'N_5+_g':{'nature':'gas','m_at':14.007,'q':5.*elec_ch},      

    #Oxygen
    'O_0+_g':{'nature':'gas','m_at':15.9994,'q':0.},
    'O_4+_g':{'nature':'gas','m_at':15.9994,'q':4.*elec_ch},
 
    #Sodium
    #    - radius = 190 pm 
    #    - collisionnal cross-section set to pi*(2.*190.*1e-12)**2.
    'Na_0+_g':{'nature':'gas','m_at':22.98976928,'sig_coll':4.536e-19,'q':0.},
    'Na_1+_g':{'nature':'gas','m_at':22.98922072,'q':elec_ch},
    
    #Magnesium
    'Mg_0+_g':{'nature':'gas','m_at':24.3050,'ion_thresh':1621.5556340659411,'q':0.},
    'Mg_1+_g':{'nature':'gas','m_at':24.3050,'q':elec_ch},

    #Silicon
    'Si_1+_g':{'nature':'gas','m_at':28.0855,'q':elec_ch},
    'Si_2+_g':{'nature':'gas','m_at':28.0855,'q':2.*elec_ch},
    'Si_3+_g':{'nature':'gas','m_at':28.0855,'q':3.*elec_ch},    

    #Calcium
    'Ca_1+_g':{'nature':'gas','m_at':40.077,'q':elec_ch},  
    
    #Iron
    'Fe_0+_g':{'nature':'gas','m_at':55.845,'q':0.},
    'Fe_11+_g':{'nature':'gas','m_at':55.845,'q':11.*elec_ch},          
    
    
    #-------------------------
    #Molecules
    #-------------------------
    
    #Dihydrogen
    #    - m_at = 2*m(H)
    #    - radius (Van der Waals) = 202 pm
    #    - collisionnal cross-section set to pi*(2.*202.*1e-12)**2.
    'H2_0+_g':{'nature':'gas','m_at':2.01588,'sig_coll':5.13e-19}, 
    
    #Water
    #    - m_at = 2*m(H) + m(O)
    #    - radius (Van der Waals) = 275 pm
    #    - collisionnal cross-section set to pi*(2.*275.*1e-12)**2.
    'H2O_0+_g':{'nature':'gas','m_at':18.01528,'sig_coll':9.50e-19},
    
    
    
    #################################################################
    #Dust grains
    #    - properties:
    # + rho : density (g cm-3)
    # + Mg: molecular mass of the gas molecules/atoms released from dust due to sublimation (amu) 
    # + L: latent heat of sublimation, in m2 s-2
    # + Pinf: vapor pressure for infinite temperature, in kg m-1 s-2
    # + alpha: evaporation coefficient that parameterises kinetic inhibition of the sublimation process (from Van Lieshout et al. 2014)
    #          set arbitrarily to 0.1 when unknown  
    #    - see change_state for how to obtain L and Pinf from A and B in Van Lieshout+2014 
    #################################################################
 
    # Alumina (gamma alumina)
    #    - from Budaj+2015
    #    - we assume sublimation releases 2 Al (Mg = 26.981539 amu) and 3 O (15.999 amu)
    'alumina':{
        'nature':'dust',
        'rho':2.9,
        'Mg':101.96, #=2*26.981539 + 3*15.999
        'L':np.nan,
        'Pinf':np.nan,
        'alpha':0.1 #arbitrary
    }, 
    # Ammonia
    #    - from Budaj+2015
    'ammonia':{   
        'nature':'dust', 
        'rho': 0.88,
        'Mg':np.nan,
        'L':np.nan,
        'Pinf':np.nan,
        'alpha':0.1 #arbitrary
    },      
    # Carbon at T=400°C 
    #    - opacity in Budaj+2015 is calculated for T=400°C
    #    - we assume the properties for Graphite in Van Lieshout et al. 2014 are valid for this type of carbon   
    'carbon0400':{
        'nature':'dust',
        'rho':1.435,            #Budaj+2015 (T=400°C). Van Lieshout et al. 2014 gives 2.16 for graphite 
        'Mg':12.011,            #Van Lieshout et al. 2014
        'L':64825562.918477856, #=93646.*k_boltz/(12.011*amu), from Van Lieshout et al. 2014
        'Pinf':8.68175420053e14,#=0.1*np.exp(36.7), from Van Lieshout et al. 2014
        'alpha':0.1             #arbitrary
    },
    # Carbon at T=1000°C 
    #    - opacity in Budaj+2015 is calculated for T=1000°C
    'carbon1000':{
        'nature':'dust',
        'rho':1.988,            #Budaj+2015 (T=400°C). Van Lieshout et al. 2014 gives 2.16 for graphite
        'Mg':12.011,            #Van Lieshout et al. 2014
        'L':64825562.918477856, #=93646.*k_boltz/(12.011*amu), from Van Lieshout et al. 2014
        'Pinf':8.68175420053e14,#=0.1*np.exp(36.7), from Van Lieshout et al. 2014
        'alpha':0.1             #arbitrary 
    },    
    # Corundum (Al2O3 = alpha alumina)
    #    - from Van Lieshout et al. 2014
    'corundum':{
        'nature':'dust',
        'rho':2.9, #from Budaj+2015, for consistency with its opacity table (see note in their section 4.1)
        'Mg':101.961,
        'L':6308798.7788943825, #=77365.*k_boltz/(101.961*amu)
        'Pinf':1.16888864240e16,#=0.1*np.exp(39.3)
        'alpha':0.1 #arbitrary
    }, 
    # Cryst. enstatite (MgSiO3)
    #    - from Van Lieshout et al. 2014
    #    - considered as an iron-free pyroxene in Budaj+2015
    'enstatite':{
        'nature':'dust',
        'rho':3.20,
        'Mg':100.389,
        'L':5707156.068029924, #=68908.*k_boltz/(100.389*amu)
        'Pinf':3.5206249346e15,#=0.1*np.exp(38.1)
        'alpha':0.1 #arbitrary
    },   
    # Cryst. fayalite (Fe2SiO4)
    #    - from Van Lieshout et al. 2014
    #    - no associated cross-section
    'fayalite':{
        'nature':'dust',
        'rho':4.39,
        'Mg':203.774,
        'L':2463536.4452695027, #=60377.*k_boltz/(203.774*amu)
        'Pinf':2.35994546824e15,#=0.1*np.exp(37.7)
        'alpha':0.1
    }, 
    # Cryst. forsterite (Mg2SiO4)
    #    - from Van Lieshout et al. 2014
    #    - considered as an iron-free olivine in Budaj+2015
    'forsterite':{
        'nature':'dust',
        'rho':3.27,
        'Mg':140.694,
        'L':3859464.3979731086, #=65308.*k_boltz/(140.694*amu)   
        'Pinf':6.44824949651e13,#=0.1*np.exp(34.1)
        'alpha':0.1 #arbitrary
    },       
    # Iron (Fe)
    #    - from Van Lieshout et al. 2014 
    'iron_dust':{
        'nature':'dust',
        'rho':7.874, 
        'Mg':55.845,  
        'L':7199201.725729473,  #=48354.*k_boltz/(55.845*amu)
#        'L':3.534e8/55.845,    #???? 
        'Pinf':4.8017425537e11, #=0.1*np.exp(29.2)
#        'Pinf':7.801e9,   ????
        'alpha':1.
    }, 
    # Olivine
    #    - Budaj+2015 opacities are for iron-enriched olivine (Mg(1) Fe(1) SiO4) with half–half of Mg–Fe atoms 
    #      Kimura+2002 values are for (Mg(1.1) Fe(0.9) SiO(4) ) olivine     
    'olmg50':{
        'nature':'dust',
        'rho': 3.0,     #Budaj+2015 
        'Mg': 169.0809, #Kimura+2002
        'L':3.21e6,     #Kimura+2002
        'Pinf':6.72e13, #Kimura+2002
        'alpha':0.1 #arbitrary
    },       
    # Perovskite (CaTiO3)
    #    - from Budaj+2015
    'perovskite':{
        'nature':'dust',
        'rho': 4.1,
        'Mg':np.nan,
        'L':np.nan,
        'Pinf':np.nan,
        'alpha':0.1      #arbitrary
    },  
    # Pyroxene 20% iron
    #    - opacity in Budaj+2015 is calculated for Mg(0.8) Fe(0.2) SiO3, but note they correspond to SiO2 alone 
    'pyrmg80':{
        'nature':'dust',
        'rho': 3.0,      #Budaj+2015  
        'Mg':106.6955,   #=0.8*24.305+0.2*55.845+28.0855+3.*15.999     
        'L':9.61e6,      #Kimura et al. 2002   
        'Pinf':3.13e10,  #Kimura et al. 2002
        'alpha':0.1      #arbitrary
    },
    # Pyroxene 60% iron
    #    - opacity in Budaj+2015 is calculated for Mg(0.4) Fe(0.6) SiO3
    #      we assume the sublimation properties from Kimura+2002 are valid for this composition, but note they correspond to SiO2 alone 
    'pyrmg40':{
        'nature':'dust',
        'rho': 3.0,      #Budaj+2015
        'Mg':119.3115,   #=0.4*24.305+0.6*55.845+28.0855+3.*15.999
        'L':5e6,     #Kimura et al. 2002
        'Pinf':3.13e10, #Kimura et al. 2002
        'alpha':0.1     #arbitrary
    }, 
    # Quartz (SiO2)
    #    - from Van Lieshout et al. 2014
    #    - no associated cross-section
    'quartz':{
        'nature':'dust',
        'rho':2.60,
        'Mg':60.084,
        'L':9609750.740329912, #=69444.*k_boltz/(60.084*amu)
        'Pinf':2.3721784213e13,#=0.1*np.exp(33.1)
        'alpha':1.0
    }, 
    # Silicon carbide (SiC)
    #    - from Van Lieshout et al. 2014
    #    - no associated cross-section
    'SiC_dust':{
        'nature':'dust',
        'rho':3.22,
        'Mg':40.10,
        'L':16268639.420874726, #=78462.*k_boltz/(40.10*amu)
        'Pinf':2.60814309975e15,#=0.1*np.exp(37.8)
        'alpha':0.1
    }, 
    # Silicon monoxide (SiO)
    #    - from Van Lieshout et al. 2014
    #    - no associated cross-section
    'SiO_dust':{
        'nature':'dust',
        'rho':2.13,
        'Mg':44.085,
        'L':9339551.536356324,  #=49520.*k_boltz/(44.085*amu)
        'Pinf':1.30187912050e13,#=0.1*np.exp(32.5)
        'alpha':0.04
    },
    # Water Ice
    #    - from Budaj+2015
    'waterice':{
        'nature':'dust',
        'rho': 1.0 ,
        'Mg':np.nan,
        'L':np.nan,
        'Pinf':np.nan,
        'alpha':0.1 #arbitrary
    },
    # Water Liquid
    #    - from Budaj+2015
    'waterliq':{
        'nature':'dust',
        'rho': 1.0 ,
        'Mg':np.nan,
        'L':np.nan,
        'Pinf':np.nan,
        'alpha':0.1 #arbitrary
    },

} 

#Add atomic mass in kg
for elem in dico_all_gases:
    if ('m_at' in dico_all_gases[elem]):dico_all_gases[elem]['mu_kg']=dico_all_gases[elem]['m_at']*amu





#################################################################
#Radiative processes
#    - each process must be uniquely associated to a given species, as indicated in the field 'species'
#    - each process is given a field 'dependance' to indicate if it depends on the local properties of the species (temperature, velocity, density, pressure) or not
#################################################################

dico_all_processes={
    
    #################################################################
    #Electronic transitions
    #    - f_os : oscillator strentgh (for radiation pressure coefficient)
    #    - delta_damping : damping wings parameter (Gamma/4 pi) in s-1
    #    - w0 : stellar line central wavelength (w in A)
    # beware of using vacuum (w0_V) or air(w0_A) reference for the transition wavelengths, depending on whether we adjust space-borne (generally vacuum-defined) or ground-based  (generally vacuum-defined) data 
    # it is possible to convert from one reference to the other, but beware that the difference vacuum - air is wavelength-dependent)
    # it is easier to select vacuum or air mode in settings, depending on the case 
    #    - use 'd' instead of a dot in transition wavelengths to avoid name issues
    #################################################################

    #Carbon 2+
    #    - see http://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=C+III&low_wl=1174&upp_wn=&upp_wl=1177&low_wn=&unit=0&submit=Retrieve+Data&de=0&java_window=3&java_mult=&format=1&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&order_out=0&max_low_enrg=&show_av=3&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on
    'CIII_1174d9':{'type':'electrans','species':'C_2+_g','w0_V':1174.93,'w0_A':None,'f_os':1.136e-01,'delta_damping':3.293e+08/(4.*pi),'dependance':True}, 
    'CIII_1175d3':{'type':'electrans','species':'C_2+_g','w0_V':1175.26,'w0_A':None,'f_os':2.724e-01,'delta_damping':4.385e+08/(4.*pi),'dependance':True}, 
    'CIII_1175d59':{'type':'electrans','species':'C_2+_g','w0_V':1175.59,'w0_A':None,'f_os':6.810e-02,'delta_damping':3.287e+08 /(4.*pi),'dependance':True}, 
    'CIII_1175d71':{'type':'electrans','species':'C_2+_g','w0_V':1175.71,'w0_A':None,'f_os':2.042e-01,'delta_damping':9.856e+08/(4.*pi),'dependance':True}, 
    'CIII_1175d99':{'type':'electrans','species':'C_2+_g','w0_V':1175.99,'w0_A':None,'f_os':9.074e-02,'delta_damping':1.313e+09/(4.*pi),'dependance':True}, 
    'CIII_1176d37':{'type':'electrans','species':'C_2+_g','w0_V':1176.37,'w0_A':None,'f_os':6.807e-02,'delta_damping':5.468e+08/(4.*pi),'dependance':True}, 

    #Neutral deuterium  
    #    - values from atomic data   
    'DI':{'type':'electrans','species':'D_0+_g','w0_V':1215.3394,'w0_A':None,'f_os':0.416,'delta_damping':0.627e9/(4.*pi),'dependance':True}, 

    #Silicon +
    #    - see atomic data and NIST
    'SiII_1197':{'type':'electrans','species':'Si_1+_g','w0_V':1197.3938,'w0_A':None,'f_os':1.50e-01,'delta_damping':1.40e+09/(4.*pi),'dependance':True}, 

    #Silicon 2+
    #    - see https://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=Si+III&limits_type=0&low_w=1205&upp_w=1216&unit=0&submit=Retrieve+Data&de=0&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&unc_out=1&order_out=0&max_low_enrg=&show_av=2&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on
    'SiIII_1206':{'type':'electrans','species':'Si_2+_g','w0_V':1206.5,'w0_A':None,'f_os':1.67,'delta_damping':2.55E+09 /(4.*pi),'dependance':True}, 

    #Lyman-alpha line
    #    - see https://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=H+I+&limits_type=0&low_w=1210&upp_w=1220&unit=0&de=0&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&unc_out=1&order_out=0&max_low_enrg=&show_av=2&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on&submit=Retrieve+Data for H I in vacuum
    'Lalpha':{'type':'electrans','species':'H_0+_g','w0_V':1215.6702,'w0_A':None,'f_os':0.41641,'delta_damping':0.627e9/(4.*pi),'dependance':True},
 
    #Oxygen 4+
    #    - see https://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=O+V&limits_type=0&low_w=1217&upp_w=1220&unit=0&submit=Retrieve+Data&de=0&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&unc_out=1&order_out=0&max_low_enrg=&show_av=2&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on
    'OV_1218':{'type':'electrans','species':'O_4+_g','w0_V':1218.344,'w0_A':None,'f_os':1.56e-06,'delta_damping':2.34e+03/(4.*pi),'dependance':True}, 
    
    #Iron 11+
    #    - see https://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=Fe+XII&limits_type=0&low_w=1240&upp_w=1244&unit=0&de=0&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&unc_out=1&order_out=0&max_low_enrg=&show_av=3&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on&submit=Retrieve+Data
    'FeXII_1241':{'type':'electrans','species':'Fe_11+_g','w0_V':1242.,'w0_A':None,'f_os':0.,'delta_damping':3.17e2 /(4.*pi),'dependance':True}, 
        
    #Nitrogen 4+
    #    - see http://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=N+V&low_wl=1238&upp_wn=&upp_wl=1244&low_wn=&unit=0&de=0&java_window=3&java_mult=&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&order_out=0&max_low_enrg=&show_av=3&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on&submit=Retrieve+Data
    'NV_1239':{'type':'electrans','species':'N_4+_g','w0_V':1238.821,'w0_A':None,'f_os':0.156,'delta_damping':0.340E+09 /(4.*pi),'dependance':True}, 
    'NV_1243':{'type':'electrans','species':'N_4+_g','w0_V':1242.804,'w0_A':None,'f_os':7.8e-2,'delta_damping':0.337E+09/(4.*pi),'dependance':True}, 

    #Nitrogen 0+
    'NI_1243':{'type':'electrans','species':'N_0+_g','w0_V':1243.179,'w0_A':None,'f_os':7.47e-02,'delta_damping':3.22e+08/(4.*pi),'dependance':True}, 

    #Carbon +
    #    - see http://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=C+II&low_wl=1334&upp_wn=&upp_wl=1336&low_wn=&unit=0&submit=Retrieve+Data&de=0&java_window=3&java_mult=&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&order_out=0&max_low_enrg=&show_av=3&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on
    'CII_1335':{'type':'electrans','species':'C_1+_g','w0_V':1334.5323,'w0_A':None,'f_os':0.128,'delta_damping':0.240E+09 /(4.*pi),'dependance':True}, 
    'CII_1336':{'type':'electrans','species':'C_1+_g','w0_V':1335.7080,'w0_A':None,'f_os':0.115,'delta_damping':0.288E+09/(4.*pi),'dependance':True}, 
                
    #Silicon 3+
    #    - see https://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=Si+IV&limits_type=0&low_w=1391&upp_w=1404&unit=0&de=0&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&unc_out=1&order_out=0&max_low_enrg=&show_av=2&max_upp_enrg=&tsb_value=1&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on&submit=Retrieve+Data
    'SiIV_1393':{'type':'electrans','species':'Si_3+_g','w0_V':1393.755,'w0_A':None,'f_os':0.513,'delta_damping':0.880E+09 /(4.*pi),'dependance':True}, 
    'SiIV_1402':{'type':'electrans','species':'Si_3+_g','w0_V':1402.770,'w0_A':None,'f_os':0.255,'delta_damping':0.863E+09/(4.*pi),'dependance':True}, 
    
    #UV doublet of ionized magnesium  
    'MgIIk':{'type':'electrans','species':'Mg_1+_g','w0_V':2796.3518,'w0_A':None,'f_os':0.615,'delta_damping':0.262e9/(4.*pi),'dependance':True}, 
    'MgIIh':{'type':'electrans','species':'Mg_1+_g','w0_V':2803.5305,'w0_A':None,'f_os':0.306,'delta_damping':0.259e9/(4.*pi),'dependance':True}, 

    #UV line of neutral magnesium      
    'MgI_2852':{'type':'electrans','species':'Mg_0+_g','w0_V':2852.9641,'w0_A':None,'f_os':1.83,'delta_damping':0.5e9/(4.*pi),'dependance':True}, 
        
    #Calcium + 
    'CaII_3933':{'type':'electrans','species':'Ca_1+_g','w0_V':3934.777,'w0_A':3933.663,'f_os':6.82e-01,'delta_damping':1.47e+08/(4.*pi),'dependance':True}, #CaIIk
    'CaII_3968':{'type':'electrans','species':'Ca_1+_g','w0_V':3969.591,'w0_A':3968.469,'f_os':3.3e-01,'delta_damping':1.4e+08/(4.*pi),'dependance':True},   #CaIIh
    
    #Visible line of neutral magnesium  
    'MgI_4572':{'type':'electrans','species':'Mg_0+_g','w0_V':4572.3767,'w0_A':None,'f_os':0.205e-5,'delta_damping':0.218e3/(4.*pi),'dependance':True}, 

    #Visible doublet of neutral sodium  
    'NaID2':{'type':'electrans','species':'Na_0+_g','w0_V':5891.583253,'w0_A':5889.95094,'f_os':0.641,'delta_damping':0.616E+08/(4.*pi),'dependance':True}, 
    'NaID1':{'type':'electrans','species':'Na_0+_g','w0_V':5897.558147,'w0_A':5895.92424,'f_os':0.320,'delta_damping':0.614E+08/(4.*pi),'dependance':True}, 
    
    #Iron lines
    #    - oscillator strengths and damping coefficient unknown are set to arbitrary values for the need of the calculation
    'FeI_5893d5':{'type':'electrans','species':'Fe_0+_g','w0_V':5893.5097,'w0_A':5891.8782,'f_os':0.1,'delta_damping':0.,'dependance':True}, 
    'FeI_5899d9':{'type':'electrans','species':'Fe_0+_g','w0_V':5899.8502,'w0_A':5898.2149,'f_os':0.1,'delta_damping':0.,'dependance':True}, 
   
    #Visible doublet of neutral potassium 
    #    - see https://physics.nist.gov/cgi-bin/ASD/lines1.pl?spectra=K+I&limits_type=0&low_w=7000&upp_w=8000&unit=0&de=0&java_window=3&java_mult=&format=0&line_out=0&en_unit=0&output=0&bibrefs=1&page_size=15&show_obs_wl=1&show_calc_wl=1&unc_out=1&order_out=0&max_low_enrg=&show_av=3&max_upp_enrg=&tsb_value=0&min_str=&A_out=0&f_out=on&intens_out=on&max_str=&allowed_out=1&forbid_out=1&min_accur=&min_intens=&conf_out=on&term_out=on&enrg_out=on&J_out=on&submit=Retrieve+Data
    'KID2':{'type':'electrans','species':'K_0+_g','w0_V':7667.008906,'w0_A':None,'f_os':0.670,'delta_damping':0.380E+08/(4.*pi),'dependance':True}, 
    'KID1':{'type':'electrans','species':'K_0+_g','w0_V':7701.083536,'w0_A':None,'f_os':0.333,'delta_damping':0.375e+08 /(4.*pi),'dependance':True}, 

    #Infrared triplet of neutral metastable Helium  
    'HeI_10832d1':{'type':'electrans','species':'He_0+_23S1','w0_V':10832.057472,'w0_A':10829.09114,'f_os':5.9902e-02,'delta_damping':1.0216e+07/(4.*pi),'dependance':True}, 
    'HeI_10833d2':{'type':'electrans','species':'He_0+_23S1','w0_V':10833.216751,'w0_A':10830.25010,'f_os':1.7974e-01,'delta_damping':1.0216e+07/(4.*pi),'dependance':True}, 
    'HeI_10833d3':{'type':'electrans','species':'He_0+_23S1','w0_V':10833.306444,'w0_A':10830.33977,'f_os':2.9958e-01,'delta_damping':1.0216e+07/(4.*pi),'dependance':True}, 
         
    #################################################################
    #Broadband processes
    #################################################################

    #Rayleigh scattering from H2
    'H2_Rayleigh':{'type':'broadband','species':'H2_0+_g','dependance':False},
    
    #Rayleigh scattering from He
    'He_Rayleigh':{'type':'broadband','species':'He_0+_g','dependance':False},

    #Hydrogen photoionization
    'H0_XUV':{'type':'broadband','species':'H_0+_g','dependance':False},


   
}  
 
